@@ -44,6 +44,9 @@ struct {
     unsigned int version;
     StringView notes;
 } constexpr version_notes[] = { {
+        20211028,
+        "» {+b}g{} and {+b}v{} do not auto-convert to lowercase anymore\n"
+    }, {
         20210828,
         "» {+u}write <filename>{} will refuse to overwrite without {+u}-force{}\n"
         "» {+u}$kak_command_fifo{} support\n"
@@ -545,11 +548,11 @@ void register_options()
                        "    terminal_status_on_top         bool\n"
                        "    terminal_set_title             bool\n"
                        "    terminal_enable_mouse          bool\n"
-                       "    terminal_change_colors         bool\n"
-                       "    terminal_wheel_up_button       int\n"
-                       "    terminal_wheel_down_button     int\n"
+                       "    terminal_synchronized          bool\n"
                        "    terminal_wheel_scroll_amount   int\n"
-                       "    terminal_shift_function_key    int\n",
+                       "    terminal_shift_function_key    int\n"
+                       "    terminal_padding_char          codepoint\n"
+                       "    terminal_padding_fill          bool\n",
                        UserInterface::Options{});
     reg.declare_option("modelinefmt", "format string used to generate the modeline",
                        "%val{bufname} %val{cursor_line}:%val{cursor_char_column} {{context_info}} {{mode_info}} - %val{client}@[%val{session}]"_str);
@@ -1249,8 +1252,12 @@ int main(int argc, char* argv[])
 }
 
 #if defined(__ELF__)
-asm(R"(
-.pushsection ".debug_gdb_scripts", "MS",@progbits,1
+#ifdef __arm__
+# define PROGBITS "%progbits"
+#else
+# define PROGBITS "@progbits"
+#endif
+asm(".pushsection \".debug_gdb_scripts\", \"MS\"," PROGBITS ",1" R"(
 .byte 4
 .ascii "kakoune-inline-gdb.py\n"
 .ascii "import os.path\n"
