@@ -223,7 +223,8 @@ void TerminalUI::Screen::set_face(const Face& face, Writer& writer)
 {
     static constexpr int fg_table[]{ 39, 30, 31, 32, 33, 34, 35, 36, 37, 90, 91, 92, 93, 94, 95, 96, 97 };
     static constexpr int bg_table[]{ 49, 40, 41, 42, 43, 44, 45, 46, 47, 100, 101, 102, 103, 104, 105, 106, 107 };
-    static constexpr int attr_table[]{ 0, 4, 7, 5, 1, 2, 3, 9 };
+    static constexpr int ul_table[]{ 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };
+    static constexpr const char* attr_table[]{ "0", "4", "4:3", "7", "5", "1", "2", "3", "9" };
 
     auto set_color = [&](bool fg, const Color& color, bool join) {
         if (join)
@@ -255,7 +256,24 @@ void TerminalUI::Screen::set_face(const Face& face, Writer& writer)
         join = true;
     }
     if (m_active_face.bg != face.bg)
+    {
         set_color(false, face.bg, join);
+        join = true;
+    }
+    if (m_active_face.underline != face.underline)
+    {
+        if (join)
+            writer.write(";");
+        if (face.underline != Color::Default)
+        {
+            if (face.underline.isRGB())
+                format_with(writer, "58:2::{}:{}:{}", face.underline.r, face.underline.g, face.underline.b);
+            else
+                format_with(writer, "58:5:{}", ul_table[(int)(char)face.underline.color]);
+        }
+        else
+            format_with(writer, "59");
+    }
     writer.write("m");
 
     m_active_face = face;
@@ -266,10 +284,6 @@ void TerminalUI::Screen::output(bool force, bool synchronized, Writer& writer)
     if (not lines)
         return;
 
-    // iTerm2 "begin synchronized update" sequence
-    if (synchronized)
-        writer.write("\033P=1s\033\\");
-
     if (force)
     {
         std::fill_n(hashes.get(), (size_t)size.line, 0);
@@ -281,77 +295,97 @@ void TerminalUI::Screen::output(bool force, bool synchronized, Writer& writer)
         return (hash_value(line.atoms) << 1) | 1; // ensure non-zero
     };
 
-    struct Change { int keep; int add; int del; };
-    Vector<Change> changes{Change{}};
-    auto new_hashes = ArrayView{lines.get(), (size_t)size.line} | transform(hash_line);
-    for_each_diff(hashes.get(), (int)size.line,
-                  new_hashes.begin(), (int)size.line,
-                  [&changes](DiffOp op, int len) mutable {
-        switch (op)
+    auto output_line = [&](const Line& line) {
+        ColumnCount pending_move = 0;
+        for (auto& [text, skip, face] : line.atoms)
         {
-            case DiffOp::Keep:
-                changes.push_back({len, 0, 0});
-                break;
-            case DiffOp::Add:
-                changes.back().add += len;
-                break;
-            case DiffOp::Remove:
-                changes.back().del += len;
-                break;
-        }
-    });
-    std::copy(new_hashes.begin(), new_hashes.end(), hashes.get());
+            if (text.empty() and skip == 0)
+                continue;
 
-    int line = 0;
-    for (auto& change : changes)
-    {
-        line += change.keep;
-        if (int del = change.del - change.add; del > 0)
-        {
-            format_with(writer, "\033[{}H\033[{}M", line + 1, del);
-            line -= del;
-        }
-        line += change.del;
-    }
-
-    line = 0;
-    for (auto& change : changes)
-    {
-        line += change.keep;
-        for (int i = 0; i < change.add; ++i)
-        {
-            if (int add = change.add - change.del; i == 0 and add > 0)
-                format_with(writer, "\033[{}H\033[{}L", line + 1, add);
-            else
-                format_with(writer, "\033[{}H", line + 1);
-
-            ColumnCount pending_move = 0;
-            for (auto& [text, skip, face] : lines[line++].atoms)
+            if (pending_move != 0)
             {
-                if (text.empty() and skip == 0)
-                    continue;
+                format_with(writer, "\033[{}C", (int)pending_move);
+                pending_move = 0;
+            }
+            set_face(face, writer);
+            writer.write(text);
+            if (skip > 3 and face.attributes == Attribute{})
+            {
+                writer.write("\033[K");
+                pending_move = skip;
+            }
+            else if (skip > 0)
+                writer.write(String{' ', skip});
+        }
+    };
 
-                if (pending_move != 0)
-                {
-                    format_with(writer, "\033[{}C", (int)pending_move);
-                    pending_move = 0;
-                }
-                set_face(face, writer);
-                writer.write(text.c_str());
-                if (skip > 3 and face.attributes == Attribute{})
-                {
-                    writer.write("\033[K");
-                    pending_move = skip;
-                }
-                else if (skip > 0)
-                    writer.write(String{' ', skip}.c_str());
+    if (synchronized)
+    {
+        writer.write("\033[?2026h"); // begin synchronized update
+
+        struct Change { int keep; int add; int del; };
+        Vector<Change> changes{Change{}};
+        auto new_hashes = ArrayView{lines.get(), (size_t)size.line} | transform(hash_line);
+        for_each_diff(hashes.get(), (int)size.line,
+                      new_hashes.begin(), (int)size.line,
+                      [&changes](DiffOp op, int len) mutable {
+            switch (op)
+            {
+                case DiffOp::Keep:
+                    changes.push_back({len, 0, 0});
+                    break;
+                case DiffOp::Add:
+                    changes.back().add += len;
+                    break;
+                case DiffOp::Remove:
+                    changes.back().del += len;
+                    break;
+            }
+        });
+        std::copy(new_hashes.begin(), new_hashes.end(), hashes.get());
+
+        int line = 0;
+        for (auto& change : changes)
+        {
+            line += change.keep;
+            if (int del = change.del - change.add; del > 0)
+            {
+                format_with(writer, "\033[{}H\033[{}M", line + 1, del);
+                line -= del;
+            }
+            line += change.del;
+        }
+
+        line = 0;
+        for (auto& change : changes)
+        {
+            line += change.keep;
+            for (int i = 0; i < change.add; ++i)
+            {
+                if (int add = change.add - change.del; i == 0 and add > 0)
+                    format_with(writer, "\033[{}H\033[{}L", line + 1, add);
+                else
+                    format_with(writer, "\033[{}H", line + 1);
+
+                output_line(lines[line++]);
             }
         }
-    }
 
-    // iTerm2 "end synchronized update" sequence
-    if (synchronized)
-        writer.write("\033P=2s\033\\");
+        writer.write("\033[?2026l"); // end synchronized update
+    }
+    else
+    {
+        for (int line = 0; line < (int)size.line; ++line)
+        {
+            auto hash = hash_line(lines[line]); 
+            if (hash == hashes[line])
+                continue;
+            hashes[line] = hash;
+
+            format_with(writer, "\033[{}H", line + 1);
+            output_line(lines[line]);
+        }
+    }
 }
 
 constexpr int TerminalUI::default_shift_function_key;
@@ -497,7 +531,7 @@ void TerminalUI::redraw(bool force)
     m_info.blit(m_screen);
 
     Writer writer{STDOUT_FILENO};
-    m_screen.output(force, m_synchronized, writer);
+    m_screen.output(force, (bool)m_synchronized, writer);
 
     auto set_cursor_pos = [&](DisplayCoord c) {
         format_with(writer, "\033[{};{}H", (int)c.line + 1, (int)c.column + 1);
@@ -774,6 +808,12 @@ Optional<Key> TerminalUI::get_next_key()
         switch (c)
         {
         case '$':
+            if (private_mode == '?' and next_char() == 'y') // DECRPM
+            {
+                if (params[0] == 2026)
+                    m_synchronized.supported = (params[0] != 0);
+                return {Key::Invalid};
+            }
             switch (params[0])
             {
             case 23: case 24:
@@ -1367,6 +1407,7 @@ void TerminalUI::setup_terminal()
         "\033[22t"    // save the current window title
         "\033[?25l"   // hide cursor
         "\033="       // set application keypad mode, so the keypad keys send unique codes
+        "\033[?2026$p" // query support for synchronize output
     );
 }
 
@@ -1429,13 +1470,14 @@ void TerminalUI::set_ui_options(const Options& options)
 
     m_status_on_top = find("terminal_status_on_top").map(to_bool).value_or(false);
     m_set_title = find("terminal_set_title").map(to_bool).value_or(true);
-    m_synchronized = find("terminal_synchronized").map(to_bool).value_or(false);
+
+    auto synchronized = find("terminal_synchronized").map(to_bool);
+    m_synchronized.set = (bool)synchronized;
+    m_synchronized.requested = synchronized.value_or(false);
 
     m_shift_function_key = find("terminal_shift_function_key").map(str_to_int_ifp).value_or(default_shift_function_key);
 
     enable_mouse(find("terminal_enable_mouse").map(to_bool).value_or(true));
-    m_wheel_up_button = find("terminal_wheel_up_button").map(str_to_int_ifp).value_or(4);
-    m_wheel_down_button = find("terminal_wheel_down_button").map(str_to_int_ifp).value_or(5);
     m_wheel_scroll_amount = find("terminal_wheel_scroll_amount").map(str_to_int_ifp).value_or(3);
 
     m_padding_char = find("terminal_padding_char").map([](StringView s) { return s.column_length() < 1 ? ' ' : s[0_char]; }).value_or(Codepoint{'~'});
